/* ivac.c -- main ivac file
 *
 * author: hackbard@hackdaworld.dyndns.org
 *         frank.zirkelbach@physik.uni-augsburg.de
 *
 * Copyright (C) 2004 Frank Zirkelbach
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

// #define USE_GTK

#include "ivac.h"

int usage(void) {

  puts("");
  puts("usage: ivac <options>");
  puts("");
  puts("options:");
  puts("-h \t\t show this help");
  puts("-n <name> \t specify your name");
  puts("-p <port> \t specify port to listen for incoming connections");
  puts("-d <device> \t specify audio device");
  puts("");

  return SUCCESS;
}

int main(int argc,char **argv) {

  /* TESTING BY NOW */

  t_ivac ivac;
  int i;

  /* default values */
  strcpy(ivac.username,"ivac");
  ivac.net.l_port=IVAC_LISTEN_PORT;
  strcpy(ivac.audio.dsp_dev,SOUND_DEVICE);
 
  /* parse argv and change default values */
  for(i=1;i<argc;i++) {
    if(argv[i][0]=='-') {
      switch(argv[i][1]) {
        case 'h':
          usage();
          return SUCCESS;
        case 'n':
          strncpy(ivac.username,argv[++i],CHAR_USERNAME);
          break;
        case 'p':
          ivac.net.l_port=atoi(argv[++i]);
          break;
        case 'd':
          strncpy(ivac.audio.dsp_dev,argv[++i],MAX_CHAR_DEVICE);
          break;
        default:
          usage();
          return ERROR;
      }
    }
    else usage();
  }

  /* clear challenger struct */
  for(i=0;i<MAX_CONNECTIONS;i++)
    memset(&(ivac.challenger[i]),0,sizeof(t_challenger));
  /* zero console buffer */
  for(i=0;i<IVAC_CONSOLE_LEN;i++)
    memset(ivac.console[i],0,IVAC_CONSOLE_STRING_LEN);

  /* set capabilities (futur: set by check routines) */
  ivac.g_cap=NETWORK;
  ivac.av_cap=AUDIO|VIDEO|DUPLEX;

  /* set event timeout */
  ivac.event.timeout.tv_sec=IVAC_S_SEC;
  ivac.event.timeout.tv_usec=IVAC_S_USEC;
  /* event init */
  event_init(&(ivac.event));

  /* input init */
  ivac.input.mode=CONTENT_BUFFER;
  input_init(&(ivac.input));

  /* network init */
  if(network_init(&(ivac.net))==N_ERROR) {
    printf("[ivac] use 'fuser -n tcp %d' to determine the process to kill!\n",
           ivac.net.l_port);
    ivac_shutdown(&ivac);
    return ERROR;
  }

  /* add listening port + stdin to (read) event system */
  event_math(ivac.net.l_fd,&(ivac.event),READ,ADD);
  event_math(0,&(ivac.event),READ,ADD);

  /* display init */
  display_init(&(ivac.display));

  /* use hardcoded audio settings by now */
  ivac.audio.fmt=BIT_8;
  ivac.audio.channels=MONO;
  ivac.audio.speed=8000;
  /* audio init */
  audio_init(&(ivac.audio));

  /* display */
  ivac_display(&(ivac.display));

  /* start event system - callbacks used: ivac_event_cb + ivac_regular_cb */
  event_start(&(ivac.event),(void *)&ivac,ivac_event_cb,ivac_regular_cb);

  return SUCCESS;
}

int ivac_shutdown(t_ivac *ivac) {

  network_shutdown(&(ivac->net));
  input_shutdown(&(ivac->input));
  event_stop(&(ivac->event));
  audio_shutdown(&(ivac->audio));
  display_shutdown(&(ivac->display));

  return SUCCESS;
}

int ivac_send_info(int channel,t_ivac *ivac) {

  char data[SEND_N_MAX];
  int size;

  size=strlen(ivac->username);

  data[0]=IVAC_SEND_NAME;
  data[1]=size;
  strncpy(data+2,ivac->username,size);
  size+=2;

  data[size]=IVAC_SEND_G_CAP;
  data[size+1]=1;
  data[size+2]=ivac->g_cap;
  size+=3;

  data[size]=IVAC_SEND_AV_CAP;
  data[size+1]=2;
  data[size+2]=(ivac->av_cap)>>8;
  data[size+3]=(ivac->av_cap)&0xff;
  size+=4;

  if(network_send(ivac->net.connection[channel].fd,data,size)==N_ERROR) {
    puts("[ivac] ivac_send_info failed");
    return ERROR;
  }

  return SUCCESS;
}

int ivac_send_quit(int channel,t_ivac *ivac) {

  char data[7]; /* one more for \0 */

  data[0]=IVAC_SEND_QUIT;
  data[1]=4;
  strcpy(data+2,"quit");

  if(network_send(ivac->net.connection[channel].fd,data,6)==N_ERROR) {
    puts("[ivac] ivac_send_quit failed");
    return ERROR;
  }

  return SUCCESS;
}

int ivac_receive_info(int channel,t_ivac *ivac) {

  char data[SEND_N_MAX];
  char c_str[IVAC_CONSOLE_STRING_LEN];
  int count,length;

  count=0;

  if((length=network_receive(ivac->net.connection[channel].fd,
                             data,SEND_N_MAX))==N_ERROR) {
    puts("[ivac] ivac_receive_info failed");
    return ERROR;
  }

  while(length-count) {
    switch(data[count]) {
      case IVAC_SEND_NAME:
        strncpy(ivac->challenger[channel].name,data+count+2,data[count+1]);
        ivac->challenger[channel].name[data[count+1]]='\0';
        count+=(data[count+1]+2);
        break;
      case IVAC_SEND_G_CAP:
        ivac->challenger[channel].g_cap=data[count+2];
        count+=3;
        break;
      case IVAC_SEND_AV_CAP:
        ivac->challenger[channel].av_cap=data[count+2]<<8;
        ivac->challenger[channel].av_cap|=data[count+3];
        count+=4;
        break;
      case IVAC_SEND_QUIT:
        if(!(strncmp(data+count+2,"quit",data[1])))
          sprintf(c_str,"channel %02d: connection closed by remote host",
                  channel);
          event_math(ivac->net.connection[channel].fd,&(ivac->event),
                     READ,REMOVE);
          network_close(&(ivac->net),channel);
          memset(&(ivac->challenger[channel]),0,sizeof(t_challenger));
          ivac_add_to_monitor(ivac,c_str);
          count+=6;
          break;
      default:
        sprintf(c_str,"ivac_receive_info, unknown character: 0x%02x\n",
               data[count]);
        ivac_add_to_monitor(ivac,c_str);
        return ERROR;
        break;
    }
  }

  return length;
}

int ivac_event_cb(t_event *event,void *ptr) {
 
  t_ivac *ivac;
  int channel;
  char c_str[IVAC_CONSOLE_STRING_LEN];

  ivac=(t_ivac *)ptr;

  /* incoming connection -- first contact => send info */
  if(FD_ISSET(ivac->net.l_fd,&(event->rfds))) {
    channel=network_manage_incoming(&(ivac->net));
    if(channel==N_E_ACCEPT)
      sprintf(c_str,"accept failed");
    else if(channel==N_E_MAXC)
      sprintf(c_str,"maximum connections reached");
    else {
      sprintf(c_str,"connection from %s port %d on channel %d",
              ivac->net.connection[channel].ip,
              ivac->net.connection[channel].port,channel);
      ivac_add_to_monitor(ivac,c_str);
      event_math(ivac->net.connection[channel].fd,event,READ,ADD);
      ivac_send_info(channel,ivac);
    }
    ivac_display_content(ivac);
  }

  /* wait for user info */
  for(channel=0;channel<MAX_CONNECTIONS;channel++) {
    if(ivac->net.connection[channel].status&C_ESTABL) {
      /* remote is sending info */
      if(FD_ISSET(ivac->net.connection[channel].fd,&(event->rfds))) {
        if(ivac_receive_info(channel,ivac)==0) {
          event_math(ivac->net.connection[channel].fd,event,READ,REMOVE);
          network_close(&(ivac->net),channel);
          sprintf(c_str,"channel %02d: broken pipe - disconnected",channel);
          ivac_add_to_monitor(ivac,c_str);
        }
        ivac_display_content(ivac);
      }
    }
  } 

  /* user interaction */
  if(FD_ISSET(0,&(event->rfds)))
    input_get_event(&(ivac->input),ivac_parse_command,ivac);

  return SUCCESS;
}

int ivac_regular_cb(t_event *event,void *ptr) {

  /* usual jobs like audio & video transmit ... */

  return SUCCESS;
}

int ivac_parse_command(t_input *input,void *ptr) {

  t_ivac *ivac;
  int i,j,k;
  int len;
  int channel;
  char *data,valid;
  char c_str[IVAC_CONSOLE_STRING_LEN];
  char arg[IVAC_ARG_COUNT][IVAC_ARG_LEN];

  ivac=(t_ivac *)ptr;
  data=input->content;
  valid=0;

  /* refresh prompt content only! */
  ivac_display_prompt_content(ivac);

  /* parse command routines */
  if(data[input->c_count-1]=='\n') {

    /* delete console string + args */
    memset(c_str,0,IVAC_CONSOLE_STRING_LEN);
    for(j=0;j<IVAC_ARG_COUNT;j++) memset(arg[j],0,IVAC_ARG_LEN);

    /* get args */
    len=0;
    while(data[len]!='\n') len++;
    i=0; j=0;
    while((i<len) && (j<IVAC_ARG_COUNT)) {
      k=0;
      while((data[i+k]!=' ')&&(data[i+k]!='\n')) {
        arg[j][k]=data[i+k];
        k++;
      }
      arg[j][k]='\0';
      j++;
      /* skip all ' ' */
      while(data[i+k]==' ') k++;
      i+=k;
    }

    /* parse command  aka arg[0] */
    if(!(strncmp(arg[0],"quit",4))) {
      valid=1;
      sprintf(c_str,"ivac shutdown ...");
      for(i=0;i<MAX_CONNECTIONS;i++) {
        if(ivac->net.connection[i].status&C_ESTABL) {
          ivac_send_quit(i,ivac);
          network_close(&(ivac->net),i);
        }
      }
      ivac_shutdown(ivac);
    }
    if(!(strncmp(arg[0],"set",3))) {
      valid=1;
      if((arg[1][0]>='0')&&(arg[1][0]<='9')) channel=atoi(arg[1]);
      else channel=-1;
      if(!(strncmp(arg[1],"name",4))) {
        strncpy(ivac->username,arg[2],CHAR_USERNAME);
        sprintf(c_str,"changed username to %s",ivac->username);
      }
      else if((channel>=0)&&(channel<MAX_CONNECTIONS)) {
        i=network_set_connection_info(&(ivac->net),channel,arg[2],atoi(arg[3]));
        if(i==N_E_IN_USE)
          sprintf(c_str,"channel %02d: connection in use",channel);
        if(i==N_SUCCESS)
          sprintf(c_str,"channel %02d: set connection info",channel);
      } else snprintf(c_str,IVAC_CONSOLE_STRING_LEN,"unknown argument: '%s'",
                      arg[1]);
    }
    if(!(strncmp(arg[0],"connect",7))) {
      valid=1;
      if((arg[1][0]>='0')&&(arg[1][0]<='9')) channel=atoi(arg[1]);
      else channel=-1;
      if((channel>=0)&&(channel<MAX_CONNECTIONS)) {
        i=network_connect(&(ivac->net),channel);
        if(i==N_E_IN_USE)
          sprintf(c_str,"channel %02d: connection in use",channel);
        else if(i==N_E_NO_INFO)
          sprintf(c_str,"channel %02d: channel not configured",channel);
        else {
          sprintf(c_str,"channel %02d: connected to %s:%d",channel,
                  ivac->net.connection[channel].ip,
                  ivac->net.connection[channel].port);
          event_math(ivac->net.connection[channel].fd,&(ivac->event),READ,ADD);
          ivac_send_info(channel,ivac);
        }
      }
      else sprintf(c_str,"invalid argument: '%s'",arg[1]);
    } 
    if(!(strncmp(arg[0],"close",5))) {
      valid=1;
      if((arg[1][0]>='0')&&(arg[1][0]<='9')) channel=atoi(arg[1]);
      else channel=-1;
      if((channel>=0)&&(channel<MAX_CONNECTIONS)) {
        if(ivac->net.connection[channel].status&C_ESTABL) {
          ivac_send_quit(channel,ivac);
          event_math(ivac->net.connection[channel].fd,&(ivac->event),
                     READ,REMOVE);
          network_close(&(ivac->net),channel);
          sprintf(c_str,"channel %02d: connection closed",channel);
          memset(&(ivac->challenger[channel]),0,sizeof(t_challenger));
        }
        else
          sprintf(c_str,"channel %02d: no active connection",channel);
      }
      else sprintf(c_str,"invalid argument: '%s'",arg[1]);
    }
    if(!(strncmp(arg[0],"select",6))) {
      valid=1;
      if((arg[1][0]>='0')&&(arg[1][0]<='9')) channel=atoi(arg[1]);
      else channel=-1;
      if((channel>=0)&&(channel<MAX_CONNECTIONS)) {
        network_select(&(ivac->net),channel);
        sprintf(c_str,"selected channel %d",channel);
      }
      else if(arg[1][0]='*') {
        network_select(&(ivac->net),MAX_CONNECTIONS);
        strcpy(c_str,"selected all channels");
      }
      else sprintf(c_str,"invalid argument: '%s'",arg[1]);
    }
    if(!(strncmp(arg[0],"deselect",8))) {
      valid=1;
      if((arg[1][0]>='0')&&(arg[1][0]<='9')) channel=atoi(arg[1]);
      else channel=-1;
      if((channel>=0)&&(channel<MAX_CONNECTIONS)) {
        network_deselect(&(ivac->net),channel);
        sprintf(c_str,"deselected channel %d",channel);
      }
      else if(arg[1][0]='*') {
        network_deselect(&(ivac->net),MAX_CONNECTIONS);
        strcpy(c_str,"deselected all channels");
      }
      else sprintf(c_str,"invalid argument: '%s'",arg[1]);
    }

    if(!valid)
      snprintf(c_str,IVAC_CONSOLE_STRING_LEN,"unknown command: '%s'",arg[0]);

    /* call network functions */
    network_manage_connection(&(ivac->net));

    /* add console string to console buffer */
    ivac_add_to_monitor(ivac,c_str);

    /* refresh whole display content */
    ivac_display_content(ivac);

    /* delete content buffer + reset counter */
    memset(input->content,0,input->c_count-1);
    input->c_count=0;

  }

  return SUCCESS;
}

int ivac_display_head(t_display *display) {

  puts("#########################################################");
  puts("##### ivac - -  Copyright (C) 2004 Frank Zirkelbach #####");
  puts("#########################################################");

  return SUCCESS;
}

int ivac_display_box(t_display *display) {
 
  return SUCCESS;
}

int ivac_display_box_content(t_ivac *ivac) {

  int channel;

  /* prepare challenger names */
  for(channel=0;channel<MAX_CONNECTIONS;channel++)
    if(ivac->challenger[channel].name[0]==0)
      strcpy(ivac->challenger[channel].name,"<empty>");

  return SUCCESS;
}

int ivac_display_console(t_display *display) {

  return SUCCESS;
}

int ivac_display_console_content(t_ivac *ivac) {

  int i;

  for(i=0;i<IVAC_CONSOLE_LEN;i++)
    printf("[ivac] console line %d: %s\n",i,ivac->console[i]);

  return SUCCESS;
}

int ivac_display_prompt(t_display *display) {

  return SUCCESS;
}

int ivac_display_prompt_content(t_ivac *ivac) {

  printf("%c",ivac->input.content[ivac->input.c_count-1]);
  fflush(NULL);

  return SUCCESS;
}

int ivac_display(t_display *display) {

  int x,y;

  /* display head */
  ivac_display_head(display);

  /* display box */
  ivac_display_box(display);

  /* display console */
  ivac_display_console(display);

  /* display command prompt */
  ivac_display_prompt(display);

  return SUCCESS;
}

int ivac_display_content(t_ivac *ivac) {

  /* display box content */
  ivac_display_box_content(ivac);

  /* display console content */
  ivac_display_console_content(ivac);

  /* display prompt content */
  ivac_display_prompt_content(ivac);

  return SUCCESS;
}

int ivac_add_to_monitor(t_ivac *ivac,char *msg) {

  int i;

  for(i=0;i<IVAC_CONSOLE_LEN-1;i++)
    memcpy(ivac->console[i],ivac->console[i+1],IVAC_CONSOLE_STRING_LEN);
  memcpy(ivac->console[IVAC_CONSOLE_LEN-1],msg,IVAC_CONSOLE_STRING_LEN);

  return SUCCESS;
}
